from json import JSONEncoder, dumps
from pathlib import Path
from typing import Any

from typeguard import typechecked


class MyEncoder(JSONEncoder):  # pragma: no cover # not directly used in code
    @typechecked
    def default(self, obj: Any) -> Any:  # pylint: disable=W0221
        if isinstance(obj, Path):
            return str(obj)
        return JSONEncoder.default(self, obj)


@typechecked
def to_json(obj: Any) -> Any:
    return dumps(obj, cls=MyEncoder)



from pathlib import Path
from random import choice
from ssl import create_default_context
from typing import Union, List, Optional, Dict, Any

from pika import ConnectionParameters, SSLOptions
from pika.credentials import PlainCredentials
from pydantic import BaseModel, validator
from typeguard import typechecked
from yaml import safe_load

from robotnikmq.error import ClientCertDoesNotExist, ClientKeyDoesNotExist, CACertDoesNotExist
from robotnikmq.error import NotConfigured
from robotnikmq.log import log


@typechecked
def _existing_file(path: Union[str, Path]) -> Path:
    try:
        path = Path(path)
        if not path.exists():
            raise ValueError(f'File path does not exist: {path}')
        return path
    except Exception as exc:
        raise ValueError(f'Invalid file path: {path}') from exc


class ServerConfig(BaseModel):
    host: str
    port: int
    user: str
    password: str
    vhost: str
    ca_cert: Path
    cert: Path
    key: Path

    _existing_ca_cert = validator('ca_cert',
                                  pre=True,
                                  always=True,
                                  allow_reuse=True)(_existing_file)
    _existing_cert = validator('cert',
                               pre=True,
                               always=True,
                               allow_reuse=True)(_existing_file)
    _existing_key = validator('key',
                              pre=True,
                              always=True,
                              allow_reuse=True)(_existing_file)

    class Config:
        json_encoders = {
            Path: str,
        }


@typechecked
def server_config(host: str, port: int, user: str, password: str, vhost: str,
                  ca_cert: Union[str, Path], cert: Union[str, Path],
                  key: Union[str, Path]) -> ServerConfig:
    ca_cert, cert, key = Path(ca_cert), Path(cert), Path(key)
    if not ca_cert.is_file():
        raise CACertDoesNotExist(ca_cert)
    if not cert.is_file():
        raise ClientCertDoesNotExist(cert)
    if not key.is_file():
        raise ClientKeyDoesNotExist(key)
    return ServerConfig(host=host, port=port, user=user, password=password,
                        vhost=vhost, ca_cert=ca_cert, cert=cert, key=key)


@typechecked
def conn_params_of(config: ServerConfig) -> ConnectionParameters:
    context = create_default_context(cafile=str(config.ca_cert))
    context.load_cert_chain(config.cert, config.key)
    return ConnectionParameters(host=config.host,
                                port=config.port,
                                virtual_host=config.vhost,
                                credentials=PlainCredentials(config.user, config.password),
                                ssl_options=SSLOptions(context, config.host))


class RobotnikConfig(BaseModel):
    tiers: List[List[ServerConfig]]

    @typechecked
    def tier(self, index: int) -> List[ServerConfig]:
        return self.tiers[index]

    @typechecked
    def a_server(self, tier: int) -> ServerConfig:
        return choice(self.tier(tier))

    @typechecked
    def as_dict(self) -> Dict[str, Any]:
        return self.dict()


@typechecked
def config_of(config_file: Optional[Path]) -> RobotnikConfig:
    if config_file is None or not config_file.exists():
        log.critical("No valid RobotnikMQ configuration file was provided")
        raise NotConfigured("No valid RobotnikMQ configuration file was provided")
    return RobotnikConfig(**safe_load(config_file.open().read()))



from time import time
from typing import Optional, Any, Dict, Union, List
from uuid import uuid4 as uuid

from pika import BasicProperties
from typeguard import typechecked

from robotnikmq.core import Robotnik, RobotnikConfig, ConnErrorCallback
from robotnikmq.rpc_server import RpcError, RpcResponse
from robotnikmq.utils import to_json


class RpcClient(Robotnik):
    @typechecked
    def __init__(self, config: Optional[RobotnikConfig] = None,
                 on_conn_error: ConnErrorCallback = None):
        super().__init__(config=config, on_conn_error=on_conn_error)
        self.response: Optional[Any] = None
        self.callback_queue = None
        self.corr_id = str(uuid())

    @typechecked
    def _on_response(self, _, __, props: BasicProperties, body: bytes) -> None:
        self.response = body.decode() if self.corr_id == props.correlation_id else None

    @typechecked
    def call(self, queue: str,
             args: Optional[Dict[str, Any]] = None,
             str_args: Optional[str] = None,
             timeout: Optional[float] = None,
             raise_on_error: bool = False) -> Union[RpcError, str, int, float,
                                                    Dict[str, Any], List[Dict[str, Any]]]:
        result = self.channel.queue_declare(queue='', exclusive=True)
        self.callback_queue = result.method.queue
        self.channel.basic_consume(
            queue=self.callback_queue,
            on_message_callback=self._on_response,
            auto_ack=True)
        self.response = None
        self.corr_id = str(uuid())
        str_args = str_args or '{}'
        self.channel.basic_publish(exchange='',
                                   routing_key=queue,
                                   properties=BasicProperties(reply_to=self.callback_queue,
                                                              correlation_id=self.corr_id),
                                   body=to_json(args) if args is not None else str_args)
        start_time = time()
        while self.response is None:
            self.connection.process_data_events()
            if timeout is not None and time() > start_time + timeout:
                raise TimeoutError(f'No response has been received for Request: {self.corr_id}')
        response = RpcResponse.from_json(self.response)
        if response is not None:
            return response.data
        error = RpcError.from_json(self.response)
        if raise_on_error:
            raise RuntimeError(error.details)
        return error



from typing import Optional, Callable

from pika.exceptions import ConnectionClosedByBroker
from typeguard import typechecked

from robotnikmq.config import RobotnikConfig
from robotnikmq.core import Robotnik, ConnErrorCallback, Message

OnMessageCallback = Callable[[Message], None]


class Subscriber(Robotnik):
    @typechecked
    def __init__(self, exchange: str,
                 binding_key: Optional[str] = None,
                 config: Optional[RobotnikConfig] = None,
                 on_conn_error: ConnErrorCallback = None):
        super().__init__(config=config, on_conn_error=on_conn_error)
        self.exchange = exchange
        self.channel.exchange_declare(exchange=self.exchange,
                                      exchange_type='topic',
                                      auto_delete=True)
        result = self.channel.queue_declare(queue='', exclusive=True)
        self.queue = result.method.queue
        self.channel.queue_bind(exchange=exchange,
                                queue=self.queue,
                                routing_key=(binding_key or '#'))

    @typechecked
    def register_listener(self, callback: OnMessageCallback) -> None:
        @typechecked
        def meta_callback(_, method, ___, body: bytes):
            callback(Message.of(body.decode()))
            self.channel.basic_ack(delivery_tag=method.delivery_tag)
        self.channel.basic_consume(queue=self.queue,
                                   on_message_callback=meta_callback,
                                   auto_ack=False)

    @typechecked
    def run(self) -> None:
        while 42:
            try:
                self.channel.start_consuming()
            except ConnectionClosedByBroker:
                pass



from collections import defaultdict
from logging import DEBUG
from typing import Dict

from logzero import LogFormatter, formatter as set_formatter
from logzero import logger as log  # pylint: disable=W0611
from typeguard import typechecked


DEFAULT_FMT = '%(color)s[%(levelname)1.1s %(asctime)s PID:%(process)d(%(processName)s)<%(threadName)s>]%(end_color)s %(message)s'
DEBUG_FMT = '%(color)s[%(levelname)1.1s %(asctime)s PID:%(process)d(%(processName)s)<%(threadName)s>%(module)s:%(funcName)s:%(lineno)d]%(end_color)s %(message)s'


class DefaultLogFormatter(LogFormatter):
    @typechecked
    def __init__(self, formatters: Dict[int, str],
                 default_format: str = DEFAULT_FMT,
                 date_format: str = '%Y-%m-%d %H:%M:%S%z'):
        self.formats = defaultdict(lambda: default_format)
        for key, value in formatters.items():
            self.formats[key] = value
        LogFormatter.__init__(self, fmt=default_format, datefmt=date_format)

    def format(self, record):
        self._fmt = self.formats[record.levelno]
        return LogFormatter.format(self, record)


set_formatter(DefaultLogFormatter({DEBUG: DEBUG_FMT}, default_format=DEFAULT_FMT))
log.debug('Debug logging configured')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['robotnikmq']

package_data = \
{'': ['*']}

install_requires = \
['arrow>=0.17.0,<0.18.0',
 'funcy>=1.15,<2.0',
 'logzero>=1.7.0,<2.0.0',
 'pika>=1.1.0,<2.0.0',
 'pydantic>=1.7.3,<2.0.0',
 'pyyaml>=5.4.1,<6.0.0',
 'typeguard>=2.1.0,<3.0.0']

setup_kwargs = {
    'name': 'robotnikmq',
    'version': '0.1.3',
    'description': 'Utilities for safe, efficient, and scalable infrastructure using RabbitMQ',
    'long_description': '# RobotnikMQ\n\nUtilities for safe, efficient, and scalable infrastructure using RabbitMQ\n\n## Usage\n\nTODO\n\n## Installation & Setup\n\nTo install robotnikmq with [`pip`](https://pip.pypa.io/en/stable/) execute the following:\n\n```bash\npip install /path/to/repo/robotnikmq\n```\n\nIf you don\'t want to re-install every time there is an update, and prefer to just pull from the git repository, then use the `-e` flag.\n\n### Configuration\n\nRobotnikMQ can be configured globally, on a per-user, or on a per-application basis. When certain functions of the RobotnikMQ library are called without a provided configuration, it will attempt to find a configuration first for the application in the current working directory `./robotnikmq.yaml`, then for the user in `~/.config/robotnikmq/robotnikmq.yaml` and then for the system in `/etc/robotnikmq/robotnikmq.yaml`. An error will be raised if a configuration is not provided and neither of those files exist.\n\nThe RobotnikMQ configuration is primarily a list of servers organized into tiers. If a given system or user can be expected to connect to the same cluster the vast majority of the time, then you can/should use a per-user or global configuration. Otherwise, simply have your application configure its own RobotnikMQ configuration (see **Usage** section).\n\nThe configuration file itself should look something like this:\n\n```yaml\ntiers:\n- - ca_cert: /home/eugene/Development/robotnikmq/tests/integration/vagrant/pki/robotnik-ca.crt\n    cert: /home/eugene/Development/robotnikmq/tests/integration/vagrant/pki/issued/rabbitmq-vm/rabbitmq-vm.crt\n    host: 127.0.0.1\n    key: /home/eugene/Development/robotnikmq/tests/integration/vagrant/pki/issued/rabbitmq-vm/rabbitmq-vm.key\n    password: \'\'\n    port: 1\n    user: \'\'\n    vhost: \'\'\n  - ca_cert: /home/eugene/Development/robotnikmq/tests/integration/vagrant/pki/robotnik-ca.crt\n    cert: /home/eugene/Development/robotnikmq/tests/integration/vagrant/pki/issued/rabbitmq-vm/rabbitmq-vm.crt\n    host: \'1\'\n    key: /home/eugene/Development/robotnikmq/tests/integration/vagrant/pki/issued/rabbitmq-vm/rabbitmq-vm.key\n    password: \'1\'\n    port: 1\n    user: \'1\'\n    vhost: \'1\'\n- - ca_cert: /home/eugene/Development/robotnikmq/tests/integration/vagrant/pki/robotnik-ca.crt\n    cert: /home/eugene/Development/robotnikmq/tests/integration/vagrant/pki/issued/rabbitmq-vm/rabbitmq-vm.crt\n    host: 127.0.0.1\n    key: /home/eugene/Development/robotnikmq/tests/integration/vagrant/pki/issued/rabbitmq-vm/rabbitmq-vm.key\n    password: hackme\n    port: 5671\n    user: robotnik\n    vhost: /robotnik\n```\n\nIn the example above, you should be able to see two tiers of servers, the first has two server configurations that are intentionally broken for testing purposes, while the second has a valid configuration (this is the configuration that is used for unit-testing).\n\nThe idea is that RobotnikMQ will first attempt to connect to all servers in the first tier in a random order, then if all of them fail, it will attempt to connect to all the servers in the second tier, and so on. This is intended to allow both load-balancing on different servers and for redundancy in case some of those servers fail. You can also configure only one tier with one server, or just a list of tiers, each of which have one server in them. This way, the secondary and tertiary servers would only be used if there is something wrong with the primary.\n\n## Development\n\n### Standards\n\n- Be excellent to each other\n- Code coverage must be at 100% for all new code, or a good reason must be provided for why a given bit of code is not covered.\n  - Example of an acceptable reason: "There is a bug in the code coverage tool and it says its missing this, but its not".\n  - Example of unacceptable reason: "This is just exception handling, its too annoying to cover it".\n- The code must pass the following analytics tools. Similar exceptions are allowable as in rule 2.\n  - `pylint --disable=C0111,W1203,R0903 --max-line-length=100 ...`\n  - `flake8 --max-line-length=100 ...`\n  - `mypy --ignore-missing-imports --follow-imports=skip --strict-optional ...`\n- All incoming information from users, clients, and configurations should be validated.\n- All internal arguments passing should be typechecked whenever possible with `typeguard.typechecked`\n\n### Development Setup\n\nUsing [poetry](https://python-poetry.org/) install from inside the repo directory:\n\n```bash\npoetry install\n```\n\nThis will set up a virtualenv which you can always activate with either `poetry shell` or run specific commands with `poetry run`. All instructions below that are meant to be run in the virtualenv will be prefaced with `(robotnikmq)$ `\n\n#### IDE Setup\n\n**Sublime Text 3**\n\n```bash\ncurl -sSL https://gitlab.com/-/snippets/2066312/raw/master/poetry.sublime-project.py | poetry run python\n```\n\n## Testing\n\nAll testing should be done with `pytest` which is installed with the `dev` requirements.\n\nTo run all the unit tests, execute the following from the repo directory:\n\n```bash\n(robotnikmq)$ pytest\n```\n\nThis should produce a coverage report in `/path/to/dewey-api/htmlcov/`\n\nWhile developing, you can use [`watchexec`](https://github.com/watchexec/watchexec) to monitor the file system for changes and re-run the tests:\n\n```bash\n(robotnikmq)$ watchexec -r -e py,yaml pytest\n```\n\nTo run a specific test file:\n\n```bash\npytest tests/unit/test_core.py\n```\n\nTo run a specific test:\n\n```bash\npytest tests/unit/test_core.py::test_hello\n```\n\nFor more information on testing, see the `pytest.ini` file as well as the [documentation](https://docs.pytest.org/en/stable/).\n\n### Vagrant Testing\n\nWhen you run `pytest` you may notice that some tests are being skipped with the message "rabbitmq-vm needs to be running", this is because those tests require other services to connect to. RobotnikMQ works primarily with RabbitMQ, as a result, a lot of the tests require a server hosting it. RobotnikMQ comes with a [vagrant testing setup](https://www.vagrantup.com/docs/cli) which can be initialized from the repo directory and executing `vagrant up --provision`. The VM comes with some basic essentials for working with RobotnikMQ. Coincidentally, the virtualized setup happens to serve as a model for how mentat can be configured.\n\n- You can log in with `vagrant ssh`\n- The VM hosts a RabbitMQ server with port-forwarding configured\n  - You can access the managment UI at: https://localhost:15671/ (username: `vagrant`, password: `Uy7ht9qbbakp5h3p2KiyyEjIX8O7VkuQ`)',
    'author': 'Eugene Kovalev',
    'author_email': 'eugene@kovalev.systems',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/robotnikmq/robotnikmq',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path
from typing import Union


class UnableToConnect(Exception):
    def __str__(self):
        return f'{self.__class__.__name__}: Robotnik is unable to connect: {self.args}'


class FileDoesNotExist(Exception):
    msg = 'Cannot find'

    def __init__(self, path: Union[Path, str]):
        self.path = path
        super().__init__()

    def __str__(self):
        return f'{self.__class__.__name__}: {self.msg} {self.path}'


class CACertDoesNotExist(FileDoesNotExist):
    msg = 'CA Certificate file not found:'


class ClientCertDoesNotExist(FileDoesNotExist):
    msg = 'Client Certificate file not found:'


class ClientKeyDoesNotExist(FileDoesNotExist):
    msg = 'Client Private Key file not found:'


class NotConfigured(Exception):
    msg = 'Robotnik is not configured'



from inspect import signature, Parameter
from json import loads as from_json
from traceback import format_exc
from socket import gethostname
from typing import Optional, Callable, Union, Any, Dict, Tuple, List
from typing import get_type_hints, get_origin, get_args
from uuid import uuid4 as uuid, UUID

from pika import BasicProperties
from pika.exceptions import AMQPError, ChannelError
from typeguard import typechecked

from robotnikmq.config import RobotnikConfig
from robotnikmq.core import Robotnik, ConnErrorCallback, thread_name, valid_json
from robotnikmq.log import log
from robotnikmq.utils import to_json


@typechecked
def _type_hint_str(typ: Any) -> str:
    if get_origin(typ) is Union:
        return f"Union[{','.join([_type_hint_str(t) for t in get_args(typ)])}]"
    return str(typ.__name__)


class RpcError():
    @typechecked
    def __init__(self, request_id: Union[str, UUID, None] = None,
                 details: Union[None, str, Dict[str, Any]] = None):
        self.request_id = request_id or uuid()
        self.details = details

    @typechecked
    def to_json(self) -> str:
        return to_json(self.to_dict())

    @typechecked
    def to_dict(self) -> Dict[str, Any]:
        return {'request_id': str(self.request_id),
                'type': 'error',
                'details': self.details}

    @staticmethod
    @typechecked
    def from_json(json_str: Union[str, bytes]) -> Optional['RpcError']:
        json_str = json_str if isinstance(json_str, str) else json_str.decode()
        log.debug(json_str)
        if valid_json(json_str):
            data = from_json(json_str)
            if all(k in data for k in {'request_id', 'type', 'details'}):
                return RpcError(request_id=data['request_id'],
                                details=data['details'])
        return None


class RpcResponse():
    @typechecked
    def __init__(self,
                 request_id: Union[str, UUID, None] = None,
                 data: Union[None, str, int, float, Dict[str, Any],
                             List[Dict[str, Any]]] = None):
        self.request_id = request_id or uuid()
        self.data = data

    @typechecked
    def to_json(self) -> str:
        return to_json(self.to_dict())

    @typechecked
    def to_dict(self) -> Dict[str, Any]:
        return {'request_id': str(self.request_id),
                'type': 'response',
                'data': self.data}

    @staticmethod
    @typechecked
    def from_json(json_str: Union[str, bytes]) -> Optional['RpcResponse']:
        json_str = json_str if isinstance(json_str, str) else json_str.decode()
        if valid_json(json_str):
            data = from_json(json_str)
            if all(k in data for k in {'request_id', 'type', 'data'}):
                return RpcResponse(request_id=data['request_id'],
                                   data=data['data'])
        return None


class RpcServer(Robotnik):
    @typechecked
    def __init__(self, config: Optional[RobotnikConfig] = None,
                 on_conn_error: ConnErrorCallback = None,
                 meta_queue_prefix: Optional[str] = None,
                 docs_queue_suffix: Optional[str] = None,
                 only_once: bool = False):
        super().__init__(config=config, on_conn_error=on_conn_error)
        self._callbacks: Dict[str, Callable] = {}
        self.meta_queue_prefix = meta_queue_prefix or gethostname()
        self.docs_queue_suffix = docs_queue_suffix or '.__doc__'
        # Typically used for testing, implies server should stop after 1 response
        self.only_once = only_once

    @typechecked
    def _register_docs(self, queue: str, callback: Callable) -> None:
        self.channel.queue_declare(queue=queue + self.docs_queue_suffix, exclusive=False)

        @typechecked
        def docs_callback(_, method, props: BasicProperties, __) -> None:
            req_id = props.correlation_id or uuid()
            response = RpcResponse(req_id,
                                   data={'rpc_queue': queue,
                                         'inputs': self._get_input_type_strings(queue),
                                         'returns': self._get_return_type_str(queue),
                                         'description': callback.__doc__})
            self.channel.basic_publish(exchange='',
                                       routing_key=props.reply_to,
                                       properties=BasicProperties(
                                        correlation_id=props.correlation_id),
                                       body=response.to_json())
            self.channel.basic_ack(delivery_tag=method.delivery_tag)

        self.channel.basic_consume(queue=queue + self.docs_queue_suffix,
                                   on_message_callback=docs_callback,
                                   auto_ack=False)

    @typechecked
    def _get_defaults(self, queue: str) -> Dict[str, Any]:
        params = signature(self._callbacks[queue]).parameters
        return {p: params[p].default for p in params if params[p].default is not Parameter.empty}

    @typechecked
    def _get_input_types(self, queue: str) -> Dict[str, Any]:
        return {k: v for k, v in get_type_hints(self._callbacks[queue]).items() if k != 'return'}

    @typechecked
    def _get_input_type_strings(self, queue: str) -> Dict[str, Any]:
        return {k: _type_hint_str(v) for k, v in
                get_type_hints(self._callbacks[queue]).items() if k != 'return'}

    @typechecked
    def _get_return_type_str(self, queue: str) -> Any:
        return _type_hint_str(get_type_hints(self._callbacks[queue])['return'])

    @staticmethod
    @typechecked
    def _is_optional(arg_type: Any) -> bool:
        return get_origin(arg_type) is Union and type(None) in get_args(arg_type)

    @staticmethod
    @typechecked
    def _valid_arg(arg_value: Any, arg_type: Any) -> bool:
        if arg_type is Any:
            return True
        if get_origin(arg_type) is Union:
            if (type(None) in get_args(arg_type)) and (arg_value is None or arg_value == {}):  # Optional
                return True
            return any(RpcServer._valid_arg(arg_value, typ) for typ in get_args(arg_type))
        if get_origin(arg_type) is dict:
            key_type, val_type = get_args(arg_type)
            return all(RpcServer._valid_arg(key, key_type) for key in arg_value.keys()) and \
                all(RpcServer._valid_arg(val, val_type) for val in arg_value.values())
        return isinstance(arg_value, arg_type)

    @typechecked
    def _valid_inputs(self, queue: str, inputs: Dict[str, Any]) -> Tuple[bool, Optional[str]]:
        inputs_with_defaults = {**self._get_defaults(queue), **inputs}
        for arg_name, arg_type in self._get_input_types(queue).items():
            if arg_name not in inputs_with_defaults and not self._is_optional(arg_type):
                return False, f'Missing required argument {arg_name}'
            if arg_name in inputs_with_defaults and not self._valid_arg(inputs_with_defaults[arg_name], arg_type):
                return False, f'Invalid type for {arg_name}'
        return True, None

    @typechecked
    def register_rpc(self, queue: str,
                     callback: Callable,
                     register_docs: bool = True) -> None:
        self.channel.queue_declare(queue=queue, exclusive=False)
        self._callbacks[queue] = callback
        if register_docs:
            self._register_docs(queue, callback)
        # TODO: servers should have an exclusive Queue for information about themselves

        @typechecked
        def meta_callback(_, method, props: BasicProperties, body: bytes):
            req_id = props.correlation_id or uuid()
            with thread_name(req_id):
                log.debug('Request received')
                try:
                    try:
                        if valid_json(body.decode()):
                            input_args: Dict[str, Any] = from_json(body.decode())
                            log.debug(f'Input JSON is valid: {input_args}')
                            valid_inputs, msg = self._valid_inputs(queue, input_args)
                            if not valid_inputs:
                                log.debug('Invalid input')
                                response = RpcError(req_id, msg).to_json()
                            elif not input_args:
                                log.debug(f'Executing: {callback}')
                                response = RpcResponse(req_id, callback()).to_json()
                            else:
                                log.debug(f'Executing: {callback} with inputs: {input_args}')
                                response = RpcResponse(req_id, callback(**input_args)).to_json()
                        else:
                            response = RpcError(req_id,
                                                'Input could not be decoded as JSON').to_json()
                    except (AMQPError, ChannelError):
                        raise  # we want this kind of exception to be caught further down
                    except Exception:  # pylint: disable=W0703
                        log.error("An error has occurred during the execution of the RPC method")
                        for line in format_exc().split('\n'):
                            log.error(line)
                        response = RpcError(request_id=req_id, details=f'There was an error '
                                                                       f'while processing the '
                                                                       f'request, please refer '
                                                                       f'to server log with '
                                                                       f'request ID: '
                                                                       f'{req_id}').to_json()
                    log.debug(f'Response: {response}')
                    self.channel.basic_publish(exchange='',
                                               routing_key=props.reply_to,
                                               properties=BasicProperties(
                                                correlation_id=props.correlation_id),
                                               body=response)
                    self.channel.basic_ack(delivery_tag=method.delivery_tag)
                    log.debug('Response sent and ack-ed')
                except (AMQPError, ChannelError):
                    log.error(f"A RabbitMQ communication error has occurred while processing "
                              f"Request ID: {req_id}")
                    for line in format_exc().split('\n'):
                        log.error(line)
            if self.only_once:
                self.channel.stop_consuming()

        self.channel.basic_consume(queue=queue,
                                   on_message_callback=meta_callback,
                                   auto_ack=False)

    @typechecked
    def run(self) -> None:
        try:
            self.channel.start_consuming()
        except KeyboardInterrupt:
            self.channel.stop_consuming()
            log.info("Shutting down server")
